import sys


PY3 = sys.version_info >= (3, 0)
if PY3:  # pragma: PY3
    import builtins
    class_types = type,
    string_types = (str, bytes)
    unicode = str
else:  # pragma: PY2
    from types import ClassType

    import __builtin__ as builtins  # noqa: F401 import unused
    class_types = (type, ClassType)
    string_types = basestring,  # noqa: F821 undefined name 'basestring'
    unicode = unicode



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for unauthorized module.
"""

import unittest

from zope.interface.verify import verifyClass

from .._compat import unicode


class UnauthorizedTests(unittest.TestCase):

    def _getTargetClass(self):
        from zExceptions.unauthorized import Unauthorized
        return Unauthorized

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_interfaces(self):
        from zope.security.interfaces import IUnauthorized

        verifyClass(IUnauthorized, self._getTargetClass())

    def test_empty(self):
        exc = self._makeOne()

        self.assertEqual(exc.name, None)
        self.assertEqual(exc.message, None)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(bytes(exc), b'Unauthorized()')
        self.assertEqual(unicode(exc), u'Unauthorized()')

    def test_ascii_message(self):
        arg = b'ERROR MESSAGE'
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, None)
        self.assertEqual(exc.message, arg)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(bytes(exc), arg)
        self.assertEqual(unicode(exc), arg.decode('ascii'))

    def test_encoded_message(self):
        arg = u'ERROR MESSAGE \u03A9'.encode('utf-8')
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, None)
        self.assertEqual(exc.message, arg)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(bytes(exc), arg)
        self.assertEqual(unicode(exc), arg.decode('utf-8'))

    def test_unicode_message(self):
        arg = u'ERROR MESSAGE \u03A9'
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, None)
        self.assertEqual(exc.message, arg)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(bytes(exc), arg.encode('utf-8'))
        self.assertEqual(unicode(exc), arg)

    def test_ascii_name(self):
        arg = b'ERROR_NAME'
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, arg)
        self.assertEqual(exc.message, None)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(
            bytes(exc),
            b"You are not allowed to access 'ERROR_NAME' in this context")
        self.assertEqual(
            unicode(exc),
            u"You are not allowed to access 'ERROR_NAME' in this context")

    def test_encoded_name(self):
        arg = u'ERROR_NAME_\u03A9'.encode('utf-8')
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, arg)
        self.assertEqual(exc.message, None)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(
            bytes(exc),
            (b"You are not allowed to access "
             b"'ERROR_NAME_\xce\xa9' in this context"))
        self.assertEqual(
            unicode(exc),
            u"You are not allowed to access "
            u"'ERROR_NAME_\u03A9' in this context")

    def test_unicode_name(self):
        arg = u'ERROR_NAME_\u03A9'
        exc = self._makeOne(arg)

        self.assertEqual(exc.name, arg)
        self.assertEqual(exc.message, None)
        self.assertEqual(exc.value, None)
        self.assertEqual(exc.needed, None)

        self.assertEqual(
            bytes(exc),
            (b"You are not allowed to access "
             b"'ERROR_NAME_\xce\xa9' in this context"))
        self.assertEqual(
            unicode(exc),
            u"You are not allowed to access "
            u"'ERROR_NAME_\u03A9' in this context")



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""General exceptions that wish they were standard exceptions

These exceptions are so general purpose that they don't belong in Zope
application-specific packages.
"""
from zope.interface import implementer
from zope.interface.common.interfaces import IException
from zope.publisher.interfaces import IBadRequest
from zope.publisher.interfaces import INotFound
from zope.publisher.interfaces import IRedirect
from zope.publisher.interfaces.http import IHTTPException
from zope.publisher.interfaces.http import IMethodNotAllowed
from zope.security.interfaces import IForbidden

from ._compat import builtins
from ._compat import class_types


status_reasons = {
    # Informational
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',

    # Successful
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    226: 'IM Used',

    # Redirection
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',

    # Client Error
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    418: "I'm a teapot",
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',
    451: 'Unavailable for Legal Reasons',

    # Server Error
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    507: 'Insufficient Storage',
    510: 'Not Extended',
    511: 'Network Authentication Required',
}


ERROR_HTML = """\
<!DOCTYPE html>
<html>
<head>
<title>Site Error</title>
<meta charset="utf-8" />
</head>
<body bgcolor="#FFFFFF">
<h2>Site Error</h2>
<p>An error was encountered while publishing this resource.
</p>
<p><strong>{title}</strong></p>

{detail}
<hr noshade="noshade"/>

<p>Troubleshooting Suggestions</p>

<ul>
<li>The URL may be incorrect.</li>
<li>The parameters passed to this resource may be incorrect.</li>
<li>A resource that this resource relies on may be
  encountering an error.</li>
</ul>

<p>If the error persists please contact the site maintainer.
Thank you for your patience.
</p>
</body></html>"""


@implementer(IHTTPException)
class HTTPException(Exception):
    body = None
    body_template = ERROR_HTML
    detail = None
    empty_body = False
    errmsg = 'Internal Server Error'
    status = 500
    title = 'Sorry, a site error occurred.'

    def setBody(self, body):
        self.body = body

    def getStatus(self):
        return self.status

    def setStatus(self, status, reason=None):
        self.status = status
        if reason is None:
            reason = status_reasons.get(status, 'Unknown')
        else:
            reason = reason
        self.errmsg = reason

    def setHeader(self, name, value):
        if not hasattr(self, 'headers'):
            self.headers = {}
        self.headers[name] = value

    def __call__(self, environ, start_response):
        headers = list(getattr(self, 'headers', {}).items())
        if not self.empty_body and \
                'content-type' not in getattr(self, 'headers', {}):
            headers.append(('content-type', 'text/html;charset=utf-8'))
        reason = status_reasons[self.getStatus()]
        start_response(
            '%d %s' % (self.getStatus(), reason),
            headers)

        if self.empty_body:
            return []

        body = self.body
        if body is None:
            message = str(self)
            detail = self.detail if self.detail is not None else message
            if self.title and detail:
                body = self.body_template.format(
                    title=self.title, detail=detail)
                body = body
            else:
                body = message
        return [body.encode('utf-8')]


class HTTPOk(HTTPException):
    """Base class for 2xx status codes."""
    errmsg = 'OK'
    status = 200


class HTTPCreated(HTTPOk):
    errmsg = 'Created'
    status = 201


class HTTPAccepted(HTTPOk):
    errmsg = 'Accepted'
    status = 202


class HTTPNonAuthoritativeInformation(HTTPOk):
    errmsg = 'Non-Authoritative Information'
    status = 203


class HTTPNoContent(HTTPOk):
    errmsg = 'No Content'
    status = 204
    empty_body = True


class HTTPResetContent(HTTPOk):
    errmsg = 'Reset Content'
    status = 205
    empty_body = True


class HTTPPartialContent(HTTPOk):
    errmsg = 'Partial Content'
    status = 206


class HTTPMultiStatus(HTTPOk):
    errmsg = 'Multi-Status'
    status = 207


class HTTPIMUsed(HTTPOk):
    errmsg = 'IM Used'
    status = 226


class HTTPRedirection(HTTPException):
    """Base class for 3xx status codes."""


class _HTTPMove(HTTPRedirection):
    """Base class for redirections requiring a location header."""

    def __init__(self, *args):
        super(_HTTPMove, self).__init__(*args)
        self.setHeader('Location', args[0])


class HTTPMultipleChoices(_HTTPMove):
    errmsg = 'Multiple Choices'
    status = 300


class HTTPMovedPermanently(_HTTPMove):
    errmsg = 'Moved Permanently'
    status = 301


@implementer(IRedirect)
class Redirect(_HTTPMove):
    errmsg = 'Found'
    status = 302


HTTPFound = Redirect  # Alias


class HTTPSeeOther(_HTTPMove):
    errmsg = 'See Other'
    status = 303


class HTTPNotModified(HTTPRedirection):
    errmsg = 'Not Modified'
    status = 304
    empty_body = True


class HTTPUseProxy(_HTTPMove):
    errmsg = 'Use Proxy'
    status = 305


class HTTPTemporaryRedirect(_HTTPMove):
    errmsg = 'Temporary Redirect'
    status = 307


class HTTPPermanentRedirect(_HTTPMove):
    errmsg = 'Permanent Redirect'
    status = 308


class HTTPError(HTTPException):
    """Base class for 4xx and 5xx status codes."""


class HTTPClientError(HTTPError):
    """Base class for 4xx status codes."""
    errmsg = 'Bad Request'
    status = 400


@implementer(IBadRequest)
class BadRequest(HTTPClientError):
    pass


HTTPBadRequest = BadRequest  # Alias


# Import cycle.
from .unauthorized import Unauthorized  # NOQA


HTTPUnauthorized = Unauthorized  # Alias


class HTTPPaymentRequired(HTTPClientError):
    errmsg = 'Payment Required'
    status = 402


@implementer(IForbidden)
class Forbidden(HTTPException):
    errmsg = 'Forbidden'
    status = 403


HTTPForbidden = Forbidden  # Alias


@implementer(INotFound)
class NotFound(HTTPException):
    errmsg = 'Not Found'
    status = 404


HTTPNotFound = NotFound  # Alias


@implementer(IMethodNotAllowed)
class MethodNotAllowed(HTTPException):
    errmsg = 'Method Not Allowed'
    status = 405


HTTPMethodNotAllowed = MethodNotAllowed  # Alias


class HTTPNotAcceptable(HTTPClientError):
    errmsg = 'Not Acceptable'
    status = 406


class HTTPProxyAuthenticationRequired(HTTPClientError):
    errmsg = 'Proxy Authentication Required'
    status = 407


class HTTPRequestTimeout(HTTPClientError):
    errmsg = 'Request Timeout'
    status = 408


class HTTPConflict(HTTPClientError):
    errmsg = 'Conflict'
    status = 409


class HTTPGone(HTTPClientError):
    errmsg = 'Gone'
    status = 410


class HTTPLengthRequired(HTTPClientError):
    errmsg = 'Length Required'
    status = 411


class HTTPPreconditionFailed(HTTPClientError):
    errmsg = 'Precondition Failed'
    status = 412


class HTTPRequestEntityTooLarge(HTTPClientError):
    errmsg = 'Request Entity Too Large'
    status = 413


class HTTPRequestURITooLong(HTTPClientError):
    errmsg = 'Request-URI Too Long'
    status = 414


class HTTPUnsupportedMediaType(HTTPClientError):
    errmsg = 'Unsupported Media Type'
    status = 415


class HTTPRequestRangeNotSatisfiable(HTTPClientError):
    errmsg = 'Request Range Not Satisfiable'
    status = 416


class HTTPExpectationFailed(HTTPClientError):
    errmsg = 'Expectation Failed'
    status = 417


class HTTPImATeapot(HTTPClientError):
    errmsg = "I'm a teapot"
    status = 418


class HTTPUnprocessableEntity(HTTPClientError):
    errmsg = 'Unprocessable Entity'
    status = 422


class ResourceLockedError(HTTPException):
    # Was defined in webdav.Lockable before.
    errmsg = 'Locked'
    status = 423


HTTPLocked = ResourceLockedError  # Alias


class HTTPFailedDependency(HTTPClientError):
    errmsg = 'Failed Dependency'
    status = 424


class HTTPUpgradeRequired(HTTPClientError):
    errmsg = 'Upgrade Required'
    status = 426


class HTTPPreconditionRequired(HTTPClientError):
    errmsg = 'Precondition Required'
    status = 428


class HTTPTooManyRequests(HTTPClientError):
    errmsg = 'Too Many Requests'
    status = 429


class HTTPRequestHeaderFieldsTooLarge(HTTPClientError):
    errmsg = 'Request Header Fields Too Large'
    status = 431


class HTTPUnavailableForLegalReasons(HTTPClientError):
    errmsg = 'Unavailable For Legal Reasons'
    status = 451


class HTTPServerError(HTTPError):
    """Base class for 5xx status codes."""
    errmsg = 'Internal Server Error'
    status = 500


@implementer(IException)
class InternalError(HTTPServerError):
    pass


HTTPInternalServerError = InternalError  # Alias


class HTTPNotImplemented(HTTPServerError):
    errmsg = 'Not Implemented'
    status = 501


class HTTPBadGateway(HTTPServerError):
    errmsg = 'Bad Gateway'
    status = 502


class HTTPServiceUnavailable(HTTPServerError):
    errmsg = 'Service Unavailable'
    status = 503


class HTTPGatewayTimeout(HTTPServerError):
    errmsg = 'Gateway Timeout'
    status = 504


class HTTPVersionNotSupported(HTTPServerError):
    errmsg = 'HTTP Version Not Supported'
    status = 505


class HTTPInsufficientStorage(HTTPServerError):
    errmsg = 'Insufficient Storage'
    status = 507


class HTTPNotExtended(HTTPServerError):
    errmsg = 'Not Extended'
    status = 510


class HTTPNetworkAuthenticationRequired(HTTPServerError):
    errmsg = 'Network Authentication Required'
    status = 511


def convertExceptionType(name):
    import zExceptions
    etype = None
    if name in builtins.__dict__:
        etype = getattr(builtins, name)
    elif hasattr(zExceptions, name):
        etype = getattr(zExceptions, name)
    if (etype is not None and
            isinstance(etype, class_types) and
            issubclass(etype, Exception)):
        return etype


def upgradeException(t, v):
    etype = convertExceptionType(t.__name__)
    if etype is not None:
        t = etype
    return t, v



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# Stock __traceback_supplement__ implementations

try:
    from html import escape
except ImportError:  # PY2
    from cgi import escape


class PathTracebackSupplement(object):
    """Implementation of ITracebackSupplement"""
    pp = None

    def __init__(self, object):
        self.object = object
        if hasattr(object, 'getPhysicalPath'):
            self.pp = '/'.join(object.getPhysicalPath())
        if hasattr(object, 'absolute_url'):
            self.source_url = '%s/manage_main' % object.absolute_url()

    def getInfo(self, as_html=0):
        if self.pp is None:
            return
        if as_html:
            return '<b>Physical Path:</b>%s' % (escape(self.pp))
        else:
            return '   - Physical Path: %s' % self.pp



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from zope.interface import implementer
from zope.security.interfaces import IUnauthorized

from zExceptions import HTTPClientError

from ._compat import PY3
from ._compat import string_types
from ._compat import unicode


@implementer(IUnauthorized)
class Unauthorized(HTTPClientError):
    """Some user wasn't allowed to access a resource
    """
    errmsg = 'Unauthorized'
    realm = None
    status = 401

    def _get_message(self):
        return self._message

    message = property(_get_message,)

    def __init__(self, message=None, value=None, needed=None,
                 name=None, realm=None, **kw):
        """Possible signatures:

        Unauthorized()
        Unauthorized(message) # Note that message includes a space
        Unauthorized(name)
        Unauthorized(name, value)
        Unauthorized(name, value, needed)
        Unauthorized(message, value, needed, name)
        Unauthorized(message, value, needed, name, realm)

        Where needed is a mapping objects with items representing requirements
        (e.g. {'permission': 'add spam'}). Any extra keyword arguments
        provides are added to needed.
        """
        if (name is None and (
                not isinstance(message, string_types) or
                len(message.split()) <= 1)):
            # First arg is a name, not a message
            name = message
            message = None

        self.name = name
        self._message = message
        self.value = value
        self.setRealm(realm)

        if kw:
            if needed:
                needed.update(kw)
            else:
                needed = kw

        self.needed = needed

    def __unicode__(self):
        if self.message is not None:
            message = (self.message if
                       isinstance(self.message, unicode) else
                       self.message.decode('utf-8'))
            return message
        if self.name is not None:
            name = (self.name if
                    isinstance(self.name, unicode) else
                    self.name.decode('utf-8'))
            return ("You are not allowed to access '%s' in this context"
                    % name)
        elif self.value is not None:
            return ("You are not allowed to access '%s' in this context"
                    % self.getValueName())
        return repr(self)

    if PY3:
        __str__ = __unicode__

        def __bytes__(self):
            return self.__unicode__().encode('utf-8')
    else:
        def __str__(self):
            return self.__unicode__().encode('utf-8')

    def setRealm(self, value):
        self.realm = value
        if value:
            self.setHeader('WWW-Authenticate', 'basic realm="%s"' % value)

    def getValueName(self):
        v = self.value
        vname = getattr(v, '__name__', None)
        if vname:
            return vname
        c = getattr(v, '__class__', type(v))
        c = getattr(c, '__name__', 'object')
        return "a particular %s" % c



import json
import unittest


class TestHTTPException(unittest.TestCase):

    def _getTargetClass(self):
        from zExceptions import HTTPException
        return HTTPException

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_body(self):
        exc = self._makeOne()
        self.assertEqual(exc.body, None)
        exc.setBody('Foo')
        self.assertEqual(exc.body, 'Foo')

    def test_headers(self):
        url = 'http://localhost/foo'
        exc = self._makeOne(url)
        self.assertTrue(getattr(exc, 'headers', None) is None)

        exc.setHeader('Location', url)
        self.assertEqual(exc.headers, {'Location': url})

    def test_status(self):
        exc = self._makeOne()
        self.assertEqual(exc.getStatus(), 500)
        self.assertEqual(exc.errmsg, 'Internal Server Error')
        exc.setStatus(503)
        self.assertEqual(exc.getStatus(), 503)
        self.assertEqual(exc.errmsg, 'Service Unavailable')

    def test_call(self):
        exc = self._makeOne('Foo Error')
        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '500 Internal Server Error',
            [('content-type', 'text/html;charset=utf-8')]
        )])
        response = b''.join(response)
        self.assertTrue(response.startswith(b'<!DOCTYPE html>'))
        self.assertTrue(b'Sorry, a site error occurred.' in response)
        self.assertTrue(b'Foo Error' in response)

    def test_call_custom(self):
        exc = self._makeOne('Foo Error')
        exc.setBody('<html>Foo</html>')
        exc.setStatus(503)

        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '503 Service Unavailable',
            [('content-type', 'text/html;charset=utf-8')]
        )])
        self.assertEqual(response, [b'<html>Foo</html>'])

    def test_call_detail(self):
        exc = self._makeOne('Foo Error')
        exc.title = 'Foo'
        exc.detail = '<p>Some foo is going on.</p>'
        exc.setStatus(503)

        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '503 Service Unavailable',
            [('content-type', 'text/html;charset=utf-8')]
        )])
        response = b''.join(response)
        self.assertTrue(response.startswith(b'<!DOCTYPE html>'))
        self.assertTrue(b'<p><strong>Foo</strong></p>' in response)
        self.assertTrue(b'<p>Some foo is going on.</p>' in response)

    def test_call_empty_body(self):
        exc = self._makeOne()
        exc.empty_body = True
        exc.setBody('Foo')
        exc.setStatus(204)

        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '204 No Content', []
        )])
        self.assertEqual(response, [])

    def test_call_extra_headers(self):
        url = 'http://localhost/foo'
        exc = self._makeOne(url)
        exc.setStatus(302)
        exc.setHeader('Location', url)

        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '302 Found',
            [('Location', url),
             ('content-type', 'text/html;charset=utf-8')]
        )])
        response = b''.join(response)
        self.assertTrue(response.startswith(b'<!DOCTYPE html>'))

    def test_custom_contenttype(self):
        url = 'http://localhost/foo'
        exc = self._makeOne(url)
        exc.setStatus(200)
        exc.setHeader('content-type', 'application/json')
        exc.setBody(json.dumps({'foo': 'bar'}))

        called = []

        def start_response(status, headers):
            called.append((status, headers))

        response = exc({'Foo': 1}, start_response)
        self.assertEqual(called, [(
            '200 OK',
            [('content-type', 'application/json')]
        )])
        response = b''.join(response)
        self.assertEqual(response, b'{"foo": "bar"}')


class TestRedirect(unittest.TestCase):

    def test_location_header_302(self):
        from zExceptions import HTTPFound
        exc = HTTPFound('/redirect_to')
        self.assertEqual(exc.headers, {'Location': '/redirect_to'})

    def test_location_header_304(self):
        from zExceptions import HTTPNotModified
        exc = HTTPNotModified('/not_modified')
        self.assertFalse(getattr(exc, 'headers', None))


class TestUnauthorized(unittest.TestCase):

    def _getTargetClass(self):
        from zExceptions import Unauthorized
        return Unauthorized

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_realm(self):
        exc = self._makeOne('message', realm='special')
        self.assertEqual(exc.headers,
                         {'WWW-Authenticate': 'basic realm="special"'})

    def test_location_header_304(self):
        from zExceptions import HTTPNotModified
        exc = HTTPNotModified('/not_modified')
        self.assertFalse(getattr(exc, 'headers', None))


class TestConvertExceptionType(unittest.TestCase):

    def _callFUT(self, name):
        from zExceptions import convertExceptionType
        return convertExceptionType(name)

    def test_name_in___builtins__(self):
        self.assertTrue(self._callFUT('SyntaxError') is SyntaxError)

    def test_name_in___builtins___not_an_exception_returns_None(self):
        self.assertTrue(self._callFUT('unichr') is None)

    def test_name_in_zExceptions(self):
        from zExceptions import Redirect
        self.assertTrue(self._callFUT('Redirect') is Redirect)

    def test_name_in_zExceptions_not_an_exception_returns_None(self):
        self.assertTrue(self._callFUT('convertExceptionType') is None)


class TestUpgradeException(unittest.TestCase):

    def _callFUT(self, t, v):
        from zExceptions import upgradeException
        return upgradeException(t, v)

    def test_non_string(self):
        t, v = self._callFUT(SyntaxError, 'TEST')
        self.assertEqual(t, SyntaxError)
        self.assertEqual(v, 'TEST')

    def test_non_string_match_by_name(self):
        class NotFound(Exception):
            pass

        t, v = self._callFUT(NotFound, 'TEST')
        from zExceptions import NotFound as zExceptions_NotFound
        self.assertIs(t, zExceptions_NotFound)
        self.assertEqual(v, 'TEST')



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
ExceptionFormatter tests.
"""

import sys
from unittest import TestCase

from zExceptions import HTTPException
from zExceptions.ExceptionFormatter import format_exception


def tb(as_html=0):
    t, v, b = sys.exc_info()
    try:
        return ''.join(format_exception(t, v, b, as_html=as_html))
    finally:
        del b


class ExceptionForTesting(HTTPException):
    pass


class TestingTracebackSupplement(object):

    source_url = '/somepath'
    line = 634
    column = 57
    warnings = ['Repent, for the end is nigh']

    def __init__(self, expression):
        self.expression = expression


class Test(TestCase):

    def testBasicNamesText(self, as_html=0):
        try:
            raise ExceptionForTesting()
        except ExceptionForTesting:
            string = tb(as_html)
            # The traceback should include the name of this function.
            self.assertIn('testBasicNamesText', string)
            # The traceback should include the name of the exception.
            self.assertIn('ExceptionForTesting', string)
        else:
            self.fail('no exception occurred')

    def testBasicNamesHTML(self):
        self.testBasicNamesText(1)

    def testSupplement(self, as_html=0):
        try:
            __traceback_supplement__ = (TestingTracebackSupplement,
                                        "You are one in a million")
            raise ExceptionForTesting
        except ExceptionForTesting:
            # import pdb; pdb.set_trace()
            string = tb(as_html)
            # The source URL
            self.assertIn('/somepath', string)
            # The line number
            self.assertIn('634', string)
            # The column number
            self.assertIn('57', string)
            # The expression
            self.assertIn("You are one in a million", string)
            # The warning
            self.assertIn("Repent, for the end is nigh", string)
        else:
            self.fail('no exception occurred')

    def testSupplementHTML(self):
        self.testSupplement(1)

    def testTracebackInfo(self, as_html=0):
        try:
            __traceback_info__ = "Adam & Eve"
            raise ExceptionForTesting
        except ExceptionForTesting:
            string = tb(as_html)
            if as_html:
                # Be sure quoting is happening.
                self.assertIn('Adam &amp; Eve', string)
            else:
                self.assertIn('Adam & Eve', string)
        else:
            self.fail('no exception occurred')

    def testTracebackInfoHTML(self):
        self.testTracebackInfo(1)

    def testMultipleLevels(self):
        # Makes sure many levels are shown in a traceback.
        def f(n):
            """Produces a (n + 1)-level traceback."""
            __traceback_info__ = 'level%d' % n
            if n > 0:
                f(n - 1)
            else:
                raise ExceptionForTesting

        try:
            f(10)
        except ExceptionForTesting:
            string = tb()
            for n in range(11):
                self.assertIn('level%d' % n, string)
        else:
            self.fail('no exception occurred')

    def testQuoteLastLine(self):
        class C:
            pass
        try:
            raise TypeError(C())
        except Exception:
            string = tb(1)
        else:
            self.fail('no exception occurred')
        self.assertIn('&lt;', string)
        self.assertIn('&gt;', string)
